Attribute VB_Name = "modMisc"
Option Explicit
'
Public Const MAX_PATH_W = 1024  ' This can actually go as high as 32767, but it's set at this to preserve a bit of memory.
'
Public Type POINTAPI
    X As Long
    Y As Long
End Type
Public Type RECT
    Left   As Long
    Top    As Long
    Right  As Long ' This is +1 (right - left = width)
    Bottom As Long ' This is +1 (bottom - top = height)
End Type
Public Type SafeArrayBound
    cElements As Long
    lLbound As Long
End Type
Public Type SafeArray2D
    cDims As Integer
    fFeatures As Integer
    cbElements As Long
    cLocks As Long
    pvData As Long
    rgSABound(0 To 1) As SafeArrayBound
End Type
Public Enum CSIDL_CODE
    CSIDL_DESKTOP = &H0&                 ' <desktop>
    CSIDL_DESKTOPDIRECTORY = &H10&       ' <user name>\Desktop
    CSIDL_APPDATA = &H1A&                ' <user name>\Application Data
    CSIDL_LOCAL_APPDATA = &H1C&          ' <user name>\Local Settings\Application Data (non roaming)
    CSIDL_SYSTEM = &H25&                 ' GetSystemDirectory()
    CSIDL_COMMON_DESKTOPDIRECTORY = &H19& ' All Users\Desktop
    CSIDL_MYDOCUMENTS = 5&
    CSIDL_MYPICTURES = &H27&
End Enum
#If False Then ' Intellisense fix.
    Public CSIDL_DESKTOP, CSIDL_DESKTOPDIRECTORY, CSIDL_APPDATA, CSIDL_LOCAL_APPDATA, CSIDL_SYSTEM, CSIDL_COMMON_DESKTOPDIRECTORY, CSIDL_MYDOCUMENTS
#End If
Public Enum ShellEnum
    Shell_HIDE = 0
    Shell_SHOWNORMAL = 1
    Shell_SHOWMINIMIZED = 2
    Shell_SHOWMAXIMIZED = 3
    Shell_RESTORE = 9
End Enum
#If False Then ' Intellisense fix.
    Public Shell_HIDE, Shell_SHOWNORMAL, Shell_SHOWMINIMIZED, Shell_SHOWMAXIMIZED
#End If
'
Private Declare Function GetCursorPos Lib "user32" (lpPoint As POINTAPI) As Long
Private Declare Function GetModuleHandle Lib "kernel32" Alias "GetModuleHandleW" (ByVal lpModuleName As Long) As Long
Private Declare Function GetUserName Lib "advapi32" Alias "GetUserNameA" (ByVal lpBuffer As String, nSize As Long) As Long
Private Declare Function GetKeyState Lib "user32" (ByVal nVirtKey As Long) As Integer
Private Declare Sub GetMem1 Lib "msvbvm60" (Source As Any, Dest As Any)
Private Declare Function SHGetPathFromIDListW Lib "shell32.dll" (ByVal pIDL As Long, ByVal pszPath As Long) As Long
Private Declare Function SHGetSpecialFolderLocation Lib "shell32" (ByVal hWndOwner As Long, ByVal nFolder As Long, pIDL As Long) As Long
Private Declare Sub CoTaskMemFree Lib "ole32.dll" (ByVal pv As Long)
Private Declare Function ShellExecuteW Lib "shell32.dll" (ByVal hWnd As Long, ByVal lpOperation As Long, ByVal lpFile As Long, ByVal lpParameters As Long, ByVal lpDirectory As Long, ByVal nShowCmd As Long) As Long
'

Public Function ShellEx(sFileName As String, Optional sOperation As String = "Open", Optional lShowCmd As ShellEnum = Shell_SHOWNORMAL) As Long
    ShellEx = ShellExecuteW(0&, StrPtr(sOperation), StrPtr(sFileName), 0&, 0&, lShowCmd)
End Function

Public Function GetSpecialFolder(ByVal CSIDL As CSIDL_CODE) As String
    ' Always include terminating backslash.
    Dim ppidl As Long
    '
    If SHGetSpecialFolderLocation(0, CSIDL, ppidl) = 0 Then
        GetSpecialFolder = String$(MAX_PATH_W, 0)
        If SHGetPathFromIDListW(ppidl, StrPtr(GetSpecialFolder)) Then
            GetSpecialFolder = RTrimNull(GetSpecialFolder)
            AddSlash GetSpecialFolder
        Else
            GetSpecialFolder = vbNullString
        End If
        CoTaskMemFree ppidl
    End If
End Function

Public Function AddSlash(sFldr As String) As String
    ' Only adds it if not there.
    ' This can be used as EITHER a sub or function, as it modifies the input.
    If Right$(sFldr, 1) <> "\" Then sFldr = sFldr & "\"
    AddSlash = sFldr
End Function

Public Function LeftMatches(sHay As String, sNeedle As String) As Boolean
    LeftMatches = Left$(sHay, Len(sNeedle)) = sNeedle
End Function

Public Function IsInCollection(coll As Collection, Key As String) As Boolean
    On Error GoTo NotInCollection
    IsObject coll.Item(Key)         ' The actual IsObject function is just used to retrieve the collection.  It doesn't matter if it's an object or not.
    IsInCollection = True
    Exit Function
    '
NotInCollection:
End Function

Public Sub SafeAddCollItem(coll As Collection, Value As Variant, Key As String)
    On Error GoTo AlreadyInCollection
    coll.Add Value, Key
AlreadyInCollection:
End Sub

Public Function UniqueFileSpec(sFileSpec As String) As String
    ' This add (#), finding a unique file.
    ' The full filespec should be provided.
    ' It MUST have an extension.
    '
    Dim sExtWithDot As String
    Dim sBase As String
    Dim i As Long
    Dim inc As Long
    '
    If Not FileOrFolderExists(sFileSpec) Then
        UniqueFileSpec = sFileSpec
        Exit Function
    End If
    '
    i = InStrRev(sFileSpec, ".")
    sExtWithDot = Mid$(sFileSpec, i)
    sBase = Left$(sFileSpec, i - 1)
    inc = 2
    Do ' No exit until we find a unique file name.
        UniqueFileSpec = sBase & " (" & Format$(inc) & ")" & sExtWithDot
        If Not FileOrFolderExists(UniqueFileSpec) Then Exit Function
        inc = inc + 1
    Loop
End Function

Public Function FileOrFolderExists(fle As String) As Boolean
    Dim i As Long
    On Error GoTo FileExistsError
    ' If no error then something existed.
    i = GetAttr(fle)
    FileOrFolderExists = True
    Exit Function
FileExistsError:
    FileOrFolderExists = False
End Function

Public Function FileExists(fle As String) As Boolean
    On Error GoTo FileExistsError
    ' If no error then something existed.
    FileExists = (GetAttr(fle) And vbDirectory) = 0
    Exit Function
FileExistsError:
    FileExists = False
End Function

Public Function FolderExists(sDir As String) As Boolean
    Dim s As String
    s = sDir
    If Right$(s, 1) = "\" Then s = Left$(s, Len(s) - 1)
    On Error GoTo FileExistsError
    ' If no error then something existed.
    FolderExists = ((GetAttr(s) And vbDirectory) = vbDirectory)
    Exit Function
FileExistsError:
    FolderExists = False
End Function

Public Function LeftMouseButtonIsDown() As Boolean
    Const VK_LBUTTON = &H1
    ' The high-order-bit (negation) says whether or not the key is down.
    ' Other bits mean different things.
    LeftMouseButtonIsDown = (GetKeyState(VK_LBUTTON) < 0)
End Function

Public Function UserName() As String
    UserName = String(201, vbNullChar)
    GetUserName UserName, 200
    UserName = RTrimNull(UserName)
End Function

Public Function SwapLongEndian(ByVal dw As Long) As Long
    ' This swaps the Byte Endianness of a Long.
    GetMem1 dw, ByVal VarPtr(SwapLongEndian) + 3
    GetMem1 ByVal VarPtr(dw) + 1, ByVal VarPtr(SwapLongEndian) + 2
    GetMem1 ByVal VarPtr(dw) + 2, ByVal VarPtr(SwapLongEndian) + 1
    GetMem1 ByVal VarPtr(dw) + 3, SwapLongEndian
End Function

Public Function ByteAlignOnWord(ByVal BitDepth As Byte, ByVal Width As Long) As Long
    ' Align any bit depth on dWord boundaries.
    ByteAlignOnWord = (((Width * BitDepth) + &H1F&) And Not &H1F&) \ &H8&
End Function

Public Function InIde6() As Boolean
    InIde6 = (GetModuleHandle(StrPtr("vba6")) <> 0)
End Function

Public Property Get MouseXtwips() As Long
    MouseXtwips = MouseX * Screen.TwipsPerPixelX
End Property

Public Property Get MouseYtwips() As Long
    MouseYtwips = MouseY * Screen.TwipsPerPixelY
End Property

Public Property Get MouseX() As Long
    Dim tP As POINTAPI
    '
    GetCursorPos tP
    MouseX = tP.X
End Property

Public Property Get MouseY() As Long
    Dim tP As POINTAPI
    '
    GetCursorPos tP
    MouseY = tP.Y
End Property

Public Function bFileExists(sFleSpec As String) As Boolean
    On Error GoTo FileExistsError
    ' If no error then something existed.
    bFileExists = (GetAttr(sFleSpec) And vbDirectory) = 0
    Exit Function
FileExistsError:
    bFileExists = False
    Exit Function
End Function

Public Function RTrimNull(s As String) As String
    Dim i As Long
    i = InStr(s, vbNullChar)
    If i Then
        RTrimNull = Left$(s, i - 1)
    Else
        RTrimNull = s
    End If
End Function

Public Property Let FrameAndChildrenEnabled(fra As Frame, bEnableStatus As Boolean)
    ' This enables/disables all child controls of a specific frame.
    Dim i As Integer
    Dim frm As Form
    '
    Set frm = fra.Parent
    For i = 0 To frm.Controls.Count - 1
        If bHasContainer(frm.Controls(i)) Then
            If frm.Controls(i).Container Is fra Then
                '
                ' Recursion to handle frames within frames.
                If TypeOf frm.Controls(i) Is Frame Then
                    FrameAndChildrenEnabled(frm.Controls(i)) = bEnableStatus
                Else
                    On Error Resume Next ' not all controls have an enabled property
                        frm.Controls(i).Enabled = bEnableStatus
                        If TypeOf frm.Controls(i) Is OptionButton Then ' Make sure Tabstop stays off for UCBED.
                            frm.Controls(i).TabStop = False
                        End If
                    On Error GoTo 0
                End If
            End If
        End If
    Next i
    fra.Enabled = bEnableStatus
End Property

Public Function bHasContainer(TheControl As Control) As Boolean
    Dim TheContainer As Control
    On Error GoTo NoContainer
    Set TheContainer = TheControl.Container
    bHasContainer = True
NoContainer:
End Function

Public Function SelectedOptCaption(opt As Object) As Variant
    ' Returns caption of selected option button, or NULL if none selected.
    Dim v As Variant
    '
    v = SelectedOptIndex(opt)
    If IsNull(v) Then
        SelectedOptCaption = Null
    Else
        SelectedOptCaption = opt(v).Caption
    End If
End Function

Public Function SelectedOptIndex(opt As Object) As Variant
    ' Returns the index number of a selected option button
    ' in an option button array.  Returns NULL if none selected.
    Dim o As OptionButton
    '
    For Each o In opt
        If o.Value Then
            SelectedOptIndex = o.Index
            Exit Function
        End If
    Next o
    SelectedOptIndex = Null
End Function

